import warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.simplefilter(action='ignore', category=Warning)

import matplotlib.pyplot as plt
import time
import os
import copy
import sys
import psutil
import numpy as np
import GPUtil

def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False
if(isnotebook()):
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm as tqdm


import torch
import torchvision
from tabulate import tabulate
from scipy.stats import logistic

################################################################################3
from monk.system.common import read_json
from monk.system.common import write_json
from monk.system.common import parse_csv
from monk.system.common import parse_csv_updated
from monk.system.common import save

from monk.system.summary import print_summary
################################################################################


################################################################################

from monk.pytorch.datasets.params import set_input_size
from monk.pytorch.datasets.params import set_batch_size
from monk.pytorch.datasets.params import set_data_shuffle
from monk.pytorch.datasets.params import set_num_processors
from monk.pytorch.datasets.params import set_weighted_sampling

from monk.pytorch.datasets.csv_dataset import DatasetCustom
from monk.pytorch.datasets.csv_dataset import DatasetCustomMultiLabel
from monk.pytorch.datasets.paths import set_dataset_train_path
from monk.pytorch.datasets.paths import set_dataset_test_path
################################################################################



################################################################################
from monk.pytorch.transforms.transforms import transform_center_crop
from monk.pytorch.transforms.transforms import transform_color_jitter
from monk.pytorch.transforms.transforms import transform_random_affine
from monk.pytorch.transforms.transforms import transform_random_crop
from monk.pytorch.transforms.transforms import transform_random_horizontal_flip
from monk.pytorch.transforms.transforms import transform_random_perspective
from monk.pytorch.transforms.transforms import transform_random_resized_crop
from monk.pytorch.transforms.transforms import transform_grayscale
from monk.pytorch.transforms.transforms import transform_random_rotation
from monk.pytorch.transforms.transforms import transform_random_vertical_flip
from monk.pytorch.transforms.transforms import transform_resize
from monk.pytorch.transforms.transforms import transform_normalize


from monk.pytorch.transforms.return_transform import set_transform_trainval
from monk.pytorch.transforms.return_transform import set_transform_test

from monk.pytorch.transforms.retrieve_transform import retrieve_trainval_transforms
from monk.pytorch.transforms.retrieve_transform import retrieve_test_transforms
################################################################################




################################################################################
from monk.system.eda.eda import class_imbalance
from monk.system.eda.eda import corrupted_missing_images
################################################################################



################################################################################
from monk.system.graphs.bar import create_plot
from monk.system.graphs.line import create_train_test_plots_accuracy
from monk.system.graphs.line import create_train_test_plots_loss
################################################################################




################################################################################
from monk.pytorch.models.layers import layer_dropout
from monk.pytorch.models.layers import layer_linear
from monk.pytorch.models.layers import activation_elu
from monk.pytorch.models.layers import activation_hardshrink
from monk.pytorch.models.layers import activation_hardtanh
from monk.pytorch.models.layers import activation_leakyrelu
from monk.pytorch.models.layers import activation_logsigmoid
from monk.pytorch.models.layers import activation_prelu
from monk.pytorch.models.layers import activation_relu
from monk.pytorch.models.layers import activation_relu6
from monk.pytorch.models.layers import activation_rrelu
from monk.pytorch.models.layers import activation_selu
from monk.pytorch.models.layers import activation_celu
from monk.pytorch.models.layers import activation_sigmoid
from monk.pytorch.models.layers import activation_softplus
from monk.pytorch.models.layers import activation_softshrink
from monk.pytorch.models.layers import activation_softsign
from monk.pytorch.models.layers import activation_tanh
from monk.pytorch.models.layers import activation_tanhshrink
from monk.pytorch.models.layers import activation_threshold
from monk.pytorch.models.layers import activation_softmin
from monk.pytorch.models.layers import activation_softmax
from monk.pytorch.models.layers import activation_logsoftmax


from monk.pytorch.models.params import set_model_name
from monk.pytorch.models.params import set_device
from monk.pytorch.models.params import set_pretrained
from monk.pytorch.models.params import set_freeze_base_network
from monk.pytorch.models.params import set_model_path


from monk.pytorch.models.common import set_parameter_requires_grad
from monk.pytorch.models.common import model_to_device
from monk.pytorch.models.common import print_grad_stats
from monk.pytorch.models.common import get_num_layers
from monk.pytorch.models.common import freeze_layers

from monk.pytorch.models.return_model import load_model
from monk.pytorch.models.return_model import setup_model
from monk.pytorch.models.return_model import debug_create_network

from monk.pytorch.models.features import CNNVisualizer
################################################################################






################################################################################
from monk.pytorch.schedulers.schedulers import scheduler_fixed
from monk.pytorch.schedulers.schedulers import scheduler_step
from monk.pytorch.schedulers.schedulers import scheduler_multistep
from monk.pytorch.schedulers.schedulers import scheduler_exponential
from monk.pytorch.schedulers.schedulers import scheduler_plateau

from monk.pytorch.schedulers.retrieve_scheduler import retrieve_scheduler

from monk.pytorch.schedulers.return_scheduler import load_scheduler
################################################################################








################################################################################
from monk.pytorch.optimizers.optimizers import adadelta
from monk.pytorch.optimizers.optimizers import adagrad
from monk.pytorch.optimizers.optimizers import adam
from monk.pytorch.optimizers.optimizers import adamw
from monk.pytorch.optimizers.optimizers import adamax
from monk.pytorch.optimizers.optimizers import rmsprop
from monk.pytorch.optimizers.optimizers import momentum_rmsprop
from monk.pytorch.optimizers.optimizers import sgd
from monk.pytorch.optimizers.optimizers import nesterov_sgd

from monk.pytorch.optimizers.retrieve_optimizer import retrieve_optimizer

from monk.pytorch.optimizers.return_optimizer import load_optimizer
################################################################################






################################################################################
from monk.pytorch.losses.losses import l1
from monk.pytorch.losses.losses import l2
from monk.pytorch.losses.losses import softmax_crossentropy
from monk.pytorch.losses.losses import crossentropy
from monk.pytorch.losses.losses import sigmoid_binary_crossentropy
from monk.pytorch.losses.losses import binary_crossentropy
from monk.pytorch.losses.losses import kldiv
from monk.pytorch.losses.losses import poisson_nll
from monk.pytorch.losses.losses import huber
from monk.pytorch.losses.losses import hinge
from monk.pytorch.losses.losses import squared_hinge
from monk.pytorch.losses.losses import multimargin
from monk.pytorch.losses.losses import squared_multimargin
from monk.pytorch.losses.losses import multilabelmargin
from monk.pytorch.losses.losses import multilabelsoftmargin

from monk.pytorch.losses.return_loss import load_loss

from monk.pytorch.losses.retrieve_loss import retrieve_loss
################################################################################






################################################################################
from monk.pytorch.training.params import set_num_epochs
from monk.pytorch.training.params import set_display_progress_realtime
from monk.pytorch.training.params import set_display_progress
from monk.pytorch.training.params import set_save_intermediate_models
from monk.pytorch.training.params import set_save_training_logs
from monk.pytorch.training.params import set_intermediate_model_prefix
################################################################################



################################################################################
from monk.pytorch.testing.process import process_single
from monk.pytorch.testing.process import process_multi
################################################################################


from monk.tf_keras_1.schedulers.imports import *
from monk.system.imports import *


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def scheduler_fixed(system_dict):
    '''
    Set learning rate fixed

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["learning_rate_scheduler"] = None;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] = "fixed";
    return system_dict;


@accepts(dict, int,  gamma=[float, int], last_epoch=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def scheduler_step(system_dict, step_size, gamma=0.1, last_epoch=-1):
    '''
    Set learning rate to decrease in regular steps

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        step_size (int): Step interval for decreasing learning rate
        gamma (str): Reduction multiplier for reducing learning rate post every step
        last_epoch (int): Set this epoch to a level post which learning rate will not be decreased

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["learning_rate_scheduler"] = "steplr";
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] = "steplr";
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["step_size"] = step_size;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["gamma"] = gamma;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["last_epoch"] = last_epoch;     
    return system_dict;


@accepts(dict, [float, int], last_epoch=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def scheduler_exponential(system_dict, gamma, last_epoch=-1):
    '''
    Set learning rate to decrease exponentially every step

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        gamma (str): Reduction multiplier for reducing learning rate post every step
        last_epoch (int): Set this epoch to a level post which learning rate will not be decreased

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["learning_rate_scheduler"] = "exponential";
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] = "exponential";
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["gamma"] = gamma;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["last_epoch"] = last_epoch;
    return system_dict;


@accepts(dict, mode=str, factor=[float, int], patience=int, verbose=bool, threshold=[float, int],
    threshold_mode=str, cooldown=int, min_lr=[float, list, int], epsilon=float, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def scheduler_plateau(system_dict, mode='min', factor=0.1, patience=10, verbose=False, threshold=0.0001, 
    threshold_mode='rel', cooldown=0, min_lr=0, epsilon=1e-08):
    '''
    Set learning rate to decrease if a metric (loss) stagnates in a plateau

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        mode (str): Either of 
                    - 'min' : lr will be reduced when the quantity monitored (loss) has stopped decreasing; 
                    - 'max' : lr reduced when the quantity monitored (accuracy) has stopped increasing. 
        factor (float): Reduction multiplier for reducing learning rate post every step
        patience (int): Number of epochs to wait before reducing learning rate
        verbose (bool): If True, all computations and wait times are printed
        threshold (float): Preset fixed to 0.0001
        threshold_mode (str): Preset fixed to 'rel' mode
        cooldown (int): Number of epochs to wait before actually applying the scheduler post the actual designated step
        min_lr (float): Set minimum learning rate, post which it will not be decreased
        epsilon (float): A small value to avoid divison by zero.
        last_epoch (int): Set this epoch to a level post which learning rate will not be decreased

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["learning_rate_scheduler"] = "reduceonplateaulr";
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] = "reduceonplateaulr";
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["mode"] = mode;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["factor"] = factor;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["patience"] = patience;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["verbose"] = verbose;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["threshold"] = threshold;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["threshold_mode"] = threshold_mode;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["cooldown"] = cooldown;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["min_lr"] = min_lr;
    system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["epsilon"] = epsilon;
    return system_dict;


import warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.simplefilter(action='ignore', category=Warning)

import matplotlib.pyplot as plt
import time
import os
import copy
import sys
import psutil
import numpy as np
import GPUtil

def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False
if(isnotebook()):
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm as tqdm


stderr = sys.stderr
sys.stderr = open(os.devnull, 'w')
import keras
sys.stderr = stderr
from keras import backend as K
import tensorflow as tf
import keras.callbacks as krc
from tabulate import tabulate

################################################################################3
from monk.system.common import read_json
from monk.system.common import write_json
from monk.system.common import parse_csv
from monk.system.common import parse_csv2
from monk.system.common import parse_csv2_updated
from monk.system.common import save

from monk.system.summary import print_summary
################################################################################





################################################################################
from monk.tf_keras_1.datasets.params import set_input_size
from monk.tf_keras_1.datasets.params import set_batch_size
from monk.tf_keras_1.datasets.params import set_data_shuffle
from monk.tf_keras_1.datasets.params import set_num_processors
from monk.tf_keras_1.datasets.params import set_weighted_sampling

from monk.tf_keras_1.datasets.paths import set_dataset_train_path
from monk.tf_keras_1.datasets.paths import set_dataset_test_path
################################################################################






################################################################################
from monk.tf_keras_1.transforms.transforms import transform_color_jitter
from monk.tf_keras_1.transforms.transforms import transform_random_affine
from monk.tf_keras_1.transforms.transforms import transform_random_horizontal_flip
from monk.tf_keras_1.transforms.transforms import transform_random_rotation
from monk.tf_keras_1.transforms.transforms import transform_random_vertical_flip
from monk.tf_keras_1.transforms.transforms import transform_mean_subtraction
from monk.tf_keras_1.transforms.transforms import transform_normalize

from monk.tf_keras_1.transforms.return_transform import set_transform_trainval
from monk.tf_keras_1.transforms.return_transform import set_transform_test
from monk.tf_keras_1.transforms.return_transform import set_transform_estimate


from monk.tf_keras_1.transforms.retrieve_transform import retrieve_trainval_transforms
from monk.tf_keras_1.transforms.retrieve_transform import retrieve_test_transforms
################################################################################




################################################################################
from monk.system.eda.eda import class_imbalance
from monk.system.eda.eda import corrupted_missing_images
################################################################################



################################################################################
from monk.system.graphs.bar import create_plot
from monk.system.graphs.line import create_train_test_plots_accuracy
from monk.system.graphs.line import create_train_test_plots_loss
################################################################################




################################################################################
from monk.tf_keras_1.models.layers import layer_dropout
from monk.tf_keras_1.models.layers import layer_linear
from monk.tf_keras_1.models.layers import layer_globalaveragepooling
from monk.tf_keras_1.models.layers import layer_flatten

from monk.tf_keras_1.models.layers import activation_leakyrelu
from monk.tf_keras_1.models.layers import activation_prelu
from monk.tf_keras_1.models.layers import activation_elu
from monk.tf_keras_1.models.layers import activation_threshold
from monk.tf_keras_1.models.layers import activation_softmax
from monk.tf_keras_1.models.layers import activation_relu
from monk.tf_keras_1.models.layers import activation_selu
from monk.tf_keras_1.models.layers import activation_softplus
from monk.tf_keras_1.models.layers import activation_softsign
from monk.tf_keras_1.models.layers import activation_tanh
from monk.tf_keras_1.models.layers import activation_sigmoid


from monk.tf_keras_1.models.params import set_model_name
from monk.tf_keras_1.models.params import set_device
from monk.tf_keras_1.models.params import set_pretrained
from monk.tf_keras_1.models.params import set_freeze_base_network
from monk.tf_keras_1.models.params import set_model_path
from monk.tf_keras_1.models.params import set_gpu_memory_fraction



from monk.tf_keras_1.models.common import set_parameter_requires_grad
from monk.tf_keras_1.models.common import print_grad_stats
from monk.tf_keras_1.models.common import get_num_layers
from monk.tf_keras_1.models.common import freeze_layers
from monk.tf_keras_1.models.common import get_num_trainable_layers
from monk.tf_keras_1.models.common import setup_device_environment

from monk.tf_keras_1.models.return_model import load_model
from monk.tf_keras_1.models.return_model import setup_model
from monk.tf_keras_1.models.return_model import debug_create_network

from monk.tf_keras_1.models.features import CNNVisualizer
################################################################################






################################################################################
from monk.tf_keras_1.schedulers.schedulers import scheduler_fixed
from monk.tf_keras_1.schedulers.schedulers import scheduler_step
from monk.tf_keras_1.schedulers.schedulers import scheduler_exponential
from monk.tf_keras_1.schedulers.schedulers import scheduler_plateau

from monk.tf_keras_1.schedulers.retrieve_scheduler import retrieve_scheduler

from monk.tf_keras_1.schedulers.return_scheduler import load_scheduler
################################################################################







################################################################################
from monk.tf_keras_1.optimizers.optimizers import adadelta
from monk.tf_keras_1.optimizers.optimizers import adagrad
from monk.tf_keras_1.optimizers.optimizers import adam
from monk.tf_keras_1.optimizers.optimizers import adamax
from monk.tf_keras_1.optimizers.optimizers import rmsprop
from monk.tf_keras_1.optimizers.optimizers import sgd
from monk.tf_keras_1.optimizers.optimizers import nesterov_sgd
from monk.tf_keras_1.optimizers.optimizers import nesterov_adam

from monk.tf_keras_1.optimizers.retrieve_optimizer import retrieve_optimizer

from monk.tf_keras_1.optimizers.return_optimizer import load_optimizer
################################################################################







################################################################################
from monk.tf_keras_1.losses.losses import l1
from monk.tf_keras_1.losses.losses import l2
from monk.tf_keras_1.losses.losses import crossentropy
from monk.tf_keras_1.losses.losses import binary_crossentropy
from monk.tf_keras_1.losses.losses import kldiv
from monk.tf_keras_1.losses.losses import hinge
from monk.tf_keras_1.losses.losses import squared_hinge

#from tf_keras_1.losses.losses import sparse_categorical_crossentropy
#from tf_keras_1.losses.losses import categorical_hinge
#from tf_keras_1.losses.losses import binary_crossentropy

from monk.tf_keras_1.losses.return_loss import load_loss

from monk.tf_keras_1.losses.retrieve_loss import retrieve_loss
################################################################################






################################################################################
from monk.tf_keras_1.training.params import set_num_epochs
from monk.tf_keras_1.training.params import set_display_progress_realtime
from monk.tf_keras_1.training.params import set_display_progress
from monk.tf_keras_1.training.params import set_save_intermediate_models
from monk.tf_keras_1.training.params import set_save_training_logs
from monk.tf_keras_1.training.params import set_intermediate_model_prefix


from monk.tf_keras_1.training.callbacks import TimeHistory
from monk.tf_keras_1.training.callbacks import MemoryHistory
################################################################################





################################################################################
from monk.tf_keras_1.testing.process import process_single
from monk.tf_keras_1.testing.process import process_multi
################################################################################



from monk.gluon.transforms.imports import *
from monk.system.imports import *



@accepts(dict, int, [tuple, list, float], [tuple, list, float], bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_random_resized_crop(system_dict, input_size, scale, ratio, train, val, test, retrieve=False):
    '''
    Apply Random Resized Cropping transformation

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        input_size (int, list): Crop size
        scale (float, tuple): scaling ratio limits; for maximum and minimum random scaling
        ratio (float, tuple): aspect ratio limits; for maximum and minmum changes to aspect ratios 
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["RandomResizedCrop"] = {};
    tmp["RandomResizedCrop"]["input_size"] = input_size;
    tmp["RandomResizedCrop"]["scale"] = scale;
    tmp["RandomResizedCrop"]["ratio"] = ratio;


    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.RandomResizedCrop(size=input_size, scale=scale, ratio=ratio));
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.RandomResizedCrop(size=input_size, scale=scale, ratio=ratio));
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.RandomResizedCrop(size=input_size, scale=scale, ratio=ratio));


    return system_dict;


@accepts(dict, int, bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_center_crop(system_dict, input_size, train, val, test, retrieve=False):
    '''
    Apply Center Cropping transformation

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        input_size (int, list): Crop size
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["CenterCrop"] = {};
    tmp["CenterCrop"]["input_size"] = input_size;

    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.CenterCrop(input_size));
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.CenterCrop(input_size));
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.CenterCrop(input_size));

    return system_dict;


@accepts(dict, [list, float, int], [list, float, int], [list, float, int], [list, float, int], bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_color_jitter(system_dict, brightness, contrast, saturation, hue, train, val, test, retrieve=False):
    '''
    Apply Color jittering transformations

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        brightness (float): Levels to jitter brightness.
                                    0 - min
                                    1 - max
        contrast (float): Levels to jitter contrast.
                                    0 - min
                                    1 - max
        saturation (float): Levels to jitter saturation.
                                    0 - min
                                    1 - max
        hue (float): Levels to jitter hue.
                                    0 - min
                                    1 - max
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["ColorJitter"] = {};
    tmp["ColorJitter"]["brightness"] = brightness;
    tmp["ColorJitter"]["contrast"] = contrast;
    tmp["ColorJitter"]["saturation"] = saturation;
    tmp["ColorJitter"]["hue"] = hue;

    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.ColorJitter(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue));
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.ColorJitter(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue));
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.ColorJitter(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue));


    return system_dict;



@accepts(dict, float, bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_random_horizontal_flip(system_dict, probability, train, val, test, retrieve=False):
    '''
    Apply random horizontal flip transformations

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        probability (float): Probability of flipping the input image
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["RandomHorizontalFlip"] = {};
    tmp["RandomHorizontalFlip"]["p"] = probability;

    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.RandomFlipLeftRight());
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.RandomFlipLeftRight());
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.RandomFlipLeftRight());


    return system_dict;


@accepts(dict, float, bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_random_vertical_flip(system_dict, probability, train, val, test, retrieve=False):
    '''
    Apply random vertical flip transformations

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        probability (float): Probability of flipping the input image
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["RandomVerticalFlip"] = {};
    tmp["RandomVerticalFlip"]["p"] = probability;

    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.RandomFlipTopBottom());
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.RandomFlipTopBottom());
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.RandomFlipTopBottom());

    return system_dict;


@accepts(dict, [float, int], bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_random_lighting(system_dict, alpha, train, val, test, retrieve=False):
    '''
    Apply random lighting transformations

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        alpha (float): > 1.0 - Randomly increase overall lighting intensity
                        <1.0 - Randomly decrease overall lighting intensity
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["RandomLighting"] = {};
    tmp["RandomLighting"]["alpha"] = alpha;

    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.RandomLighting(alpha));
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.RandomLighting(alpha));
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.RandomLighting(alpha));


    return system_dict;


@accepts(dict, int, bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_resize(system_dict, input_size, train, val, test, retrieve=False):
    '''
    Apply standard resizing

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        input_size (int, list): expected final size
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["Resize"] = {};
    tmp["Resize"]["input_size"] = input_size;

    if(type(input_size) == tuple or type(input_size) == list):
        h = input_size[0];
        w = input_size[1];
    else:
        h = input_size;
        w = input_size;

    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.Resize(size=(w, h)));
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.Resize(size=(w, h)));
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.Resize(size=(w, h)));

    return system_dict;


@accepts(dict, [float, list], [float, list], bool, bool, bool, retrieve=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def transform_normalize(system_dict, mean, std, train, val, test, retrieve=False):
    '''
    Apply mean subtraction and standard normalization

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        mean (float, list): Mean value for subtraction
        std (float, list): Normalization factor
        train (bool): If True, transform applied to training data
        val (bool): If True, transform applied to validation data
        test (bool): If True, transform applied to testing/inferencing data

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["Normalize"] = {};
    tmp["Normalize"]["mean"] = mean;
    tmp["Normalize"]["std"] = std;
    system_dict["local"]["normalize"] = True;
    input_size = system_dict["dataset"]["params"]["input_size"];

    if(type(system_dict["dataset"]["params"]["input_size"]) == tuple or type(system_dict["dataset"]["params"]["input_size"]) == list):
        h = system_dict["dataset"]["params"]["input_size"][0];
        w = system_dict["dataset"]["params"]["input_size"][1];
    else:
        h = system_dict["dataset"]["params"]["input_size"];
        w = system_dict["dataset"]["params"]["input_size"];

    if(train):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["train"].append(tmp);
        system_dict["local"]["transforms_train"].append(transforms.Resize(size=(w, h)));
        system_dict["local"]["transforms_train"].append(transforms.ToTensor())
        system_dict["local"]["transforms_train"].append(transforms.Normalize(mean=mean, std=std));
    if(val):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["val"].append(tmp);
        system_dict["local"]["transforms_val"].append(transforms.Resize(size=(w, h)));
        system_dict["local"]["transforms_val"].append(transforms.ToTensor())
        system_dict["local"]["transforms_val"].append(transforms.Normalize(mean=mean, std=std));
    if(test):
        if(not retrieve):
            system_dict["dataset"]["transforms"]["test"].append(tmp);
        system_dict["local"]["transforms_test"].append(transforms.Resize(size=(w, h)));
        system_dict["local"]["transforms_test"].append(transforms.ToTensor())
        system_dict["local"]["transforms_test"].append(transforms.Normalize(mean=mean, std=std));

    return system_dict;


from monk.tf_keras_1.transforms.imports import *
from monk.tf_keras_1.transforms.common import set_transforms
from monk.system.imports import *


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def retrieve_trainval_transforms(system_dict):
    '''
    Retrieve training and validdation transforms in copy-from, and resume states

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    set_phases = ["train", "val"];
    system_dict = set_transforms(system_dict, set_phases);
    return system_dict;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def retrieve_test_transforms(system_dict):
    '''
    Retrieve testing transforms in copy-from, and resume states

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    set_phases = ["test"];
    system_dict= set_transforms(system_dict, set_phases);
    return system_dict;


import os
import sys

import psutil

from monk.gluon_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_functionality_tests.gluon.common import print_start
from monk.pip_functionality_tests.gluon.common import print_status



def test_switch_expert(system_dict):
    forward = True;
    if(not os.path.isdir("datasets")):
        os.system("! wget --load-cookies /tmp/cookies.txt \"https://docs.google.com/uc?export=download&confirm=$(wget --save-cookies /tmp/cookies.txt --keep-session-cookies --no-check-certificate 'https://docs.google.com/uc?export=download&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2' -O- | sed -rn 's/.*confirm=([0-9A-Za-z_]+).*/\1\n/p')&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2\" -O datasets.zip && rm -rf /tmp/cookies.txt")
        os.system("! unzip -qq datasets.zip")

    test = "switch_expert_object_creation";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf = prototype(verbose=0);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "switch_expert_Prototype()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Prototype("sample-project-1", "sample-experiment-6");
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");



    test = "switch_expert_switch_mode()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Switch_Mode(eval_infer=True)
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");




    test = "switch_expert_Model_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Model_Params(model_path = ["workspace/sample-project-1/sample-experiment-5/output/models/intermediate_model_-symbol.json", 
               "workspace/sample-project-1/sample-experiment-5/output/models/intermediate_model_-0000.params"],
               use_gpu=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "switch_expert_Model()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Model();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "switch_expert_update_input_size()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.update_input_size(224);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");



    
    test = "switch_expert_Infer-Img()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            img_name = "datasets/dataset_cats_dogs_test/0.jpg";
            predictions = gtf.Infer(img_name=img_name, return_raw=True);
            img_name = "datasets/dataset_cats_dogs_test/84.jpg";
            predictions = gtf.Infer(img_name=img_name, return_raw=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    
    test = "switch_expert_Infer-Folder()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            inference_dataset = "datasets/dataset_cats_dogs_test/";
            output = gtf.Infer(img_dir=inference_dataset, return_raw=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    

    test = "switch_expert_Dataset_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Dataset_Params(dataset_path="datasets/dataset_cats_dogs_eval", input_size=224);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");




    test = "switch_expert_Dataset()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Dataset();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "switch_expert_Evaluate()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            accuracy, class_based_accuracy = gtf.Evaluate();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    

    test = "switch_expert_switch_mode()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Switch_Mode(train=True)
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");



    test = "expert_train_Dataset_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Dataset_Params(dataset_path="datasets/dataset_cats_dogs_train",
                split=0.9, input_size=224, batch_size=16, shuffle_data=True, num_processors=3);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "expert_train_apply_transforms()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.apply_random_vertical_flip(train=True, val=True);
            gtf.apply_random_horizontal_flip(train=True, val=True);
            gtf.apply_random_lighting(train=True, val=True);
            gtf.apply_normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225], train=True, val=True, test=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "expert_train_Dataset()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Dataset();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "expert_train_Model_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Model_Params(model_name="resnet18_v1", freeze_base_network=True, use_gpu=True, use_pretrained=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");
      

    test = "expert_train_Model()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Model();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 


    test = "expert_train_lr_multistep_decrease()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.lr_multistep_decrease([1, 3], gamma=0.9);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 


    test = "expert_train_optimizer_sgd()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.optimizer_sgd(0.001, momentum=0.9);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 



    test = "expert_train_loss_softmax_crossentropy()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.loss_softmax_crossentropy();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 

    
    test = "expert_train_Training_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Training_Params(num_epochs=4, display_progress=True, display_progress_realtime=True, 
                save_intermediate_models=True, intermediate_model_prefix="intermediate_model_", save_training_logs=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");



    test = "expert_train_Train()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf.Train();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");



    
    return system_dict


from monk.tf_keras_1.finetune.imports import *
from monk.system.imports import *
from monk.tf_keras_1.finetune.level_1_dataset_base import finetune_dataset


class finetune_model(finetune_dataset):
    '''
    Base class for Model setup

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @accepts("self", path=[bool, str, list], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_model_final(self, path=False):
        '''
        Setup model based on set parameters

        Args:
            path (str): Dummy variable

        Returns:
            None
        '''
        self.custom_print("Model Details");
        if(self.system_dict["model"]["params"]["model_path"]):
            if(os.path.isfile(self.system_dict["model"]["params"]["model_path"])):
                self.custom_print("    Loading model - {}".format(self.system_dict["model"]["params"]["model_path"]));
                self.system_dict = setup_device_environment(self.system_dict);
                self.system_dict["local"]["model"] = load_model(self.system_dict, external_path=self.system_dict["model"]["params"]["model_path"]);
                self.custom_print("    Model loaded!");
                self.custom_print("");

                self.system_dict = get_num_layers(self.system_dict);

                self.system_dict = get_num_trainable_layers(self.system_dict);
                self.system_dict["model"]["status"] = True;
            else:
                msg = "Model not found - {}\n".format(self.system_dict["model"]["params"]["model_path"]);
                msg += "Previous Training Incomplete.";
                raise ConstraintError(msg);  

              
        elif(self.system_dict["states"]["eval_infer"]):
            if(os.path.isfile(self.system_dict["model_dir_relative"] + 'final.h5')):
                self.custom_print("    Loading model - {}".format(self.system_dict["model_dir_relative"] + 'final.h5'));
                self.system_dict = setup_device_environment(self.system_dict);
                self.system_dict["local"]["model"] = load_model(self.system_dict, final=True);
                self.custom_print("    Model loaded!");
                self.custom_print("");
            else:
                msg = "Model not found - {}\n".format(self.system_dict["model_dir_relative"] + 'final.h5');
                msg += "Previous Training Incomplete.";
                raise ConstraintError(msg);

        else:
            if(self.system_dict["states"]["resume_train"]):
                if(os.path.isfile(self.system_dict["model_dir_relative"] + 'resume_state.h5')):
                    self.custom_print("    Loading model - {}".format(self.system_dict["model_dir_relative"] + 'resume_state.h5'));
                    self.system_dict = setup_device_environment(self.system_dict);
                    self.system_dict["local"]["model"] = load_model(self.system_dict, resume=True);
                else:
                    msg = "Model not found - \"{}\"\n".format(self.system_dict["model_dir_relative"] + 'resume_state.h5');
                    msg += "Training not started. Cannot Run resume Mode";
                    raise ConstraintError(msg);
            
            elif(self.system_dict["states"]["copy_from"]):
                model_path = self.system_dict["master_systems_dir_relative"] + self.system_dict["origin"][0] + "/" + self.system_dict["origin"][1] + "/output/models/";
                if(os.path.isfile(model_path + 'final.h5')):
                    self.custom_print("    Loading model - {}".format(model_path + 'final.h5'));
                    self.system_dict = setup_device_environment(self.system_dict);
                    self.system_dict["local"]["model"] = load_model(self.system_dict, path=model_path, final=True);
                else:
                    msg = "Model not found - {}\n".format(model_path + "final.h5");
                    msg += "Previous Training Incomplete.";
                    raise ConstraintError(msg);

            else:
                self.custom_print("    Loading pretrained model");
                self.system_dict = setup_device_environment(self.system_dict);
                self.system_dict = setup_model(self.system_dict);

            
            self.custom_print("    Model Loaded on device");

            self.system_dict = get_num_layers(self.system_dict);

            self.system_dict = get_num_trainable_layers(self.system_dict);
            self.system_dict["model"]["status"] = True;


            if(self.system_dict["model"]["type"] == "custom"):
                self.custom_print("        Model name:                           {}".format("Custom Model"));
            else:
                self.custom_print("        Model name:                           {}".format(self.system_dict["model"]["params"]["model_name"]));
            self.custom_print("        Num layers in model:  {}".format(self.system_dict["model"]["params"]["num_layers"]));
            self.custom_print("        Num trainable layers: {}".format(self.system_dict["model"]["params"]["num_params_to_update"]));
            self.custom_print("");
            
    ###############################################################################################################################################








from monk.gluon.transforms.imports import *
from monk.system.imports import *



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_transform_trainval(system_dict):
    '''
    Load training and validation transforms in main state

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    if(not system_dict["local"]["normalize"]):
        if(not system_dict["local"]["applied_train_tensor"]):
            if(type(system_dict["dataset"]["params"]["input_size"]) == tuple or type(system_dict["dataset"]["params"]["input_size"]) == list):
                h = system_dict["dataset"]["params"]["input_size"][0];
                w = system_dict["dataset"]["params"]["input_size"][1];
            else:
                h = system_dict["dataset"]["params"]["input_size"];
                w = system_dict["dataset"]["params"]["input_size"];

            system_dict["local"]["transforms_train"].append(transforms.Resize(size=(w, h)));
            system_dict["local"]["transforms_train"].append(transforms.ToTensor());
            system_dict["local"]["transforms_val"].append(transforms.Resize(size=(w, h)));
            system_dict["local"]["transforms_val"].append(transforms.ToTensor());
            system_dict["local"]["applied_train_tensor"] = True;
    system_dict["local"]["data_transforms"]["train"] = transforms.Compose(system_dict["local"]["transforms_train"]);
    system_dict["local"]["data_transforms"]["val"] = transforms.Compose(system_dict["local"]["transforms_val"]);

    return system_dict;



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_transform_test(system_dict):
    '''
    Load testing transforms in main state

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    if(not system_dict["local"]["normalize"]):
        if(not system_dict["local"]["applied_test_tensor"]):

            if(type(system_dict["dataset"]["params"]["input_size"]) == tuple or type(system_dict["dataset"]["params"]["input_size"]) == list):
                h = system_dict["dataset"]["params"]["input_size"][0];
                w = system_dict["dataset"]["params"]["input_size"][1];
            else:
                h = system_dict["dataset"]["params"]["input_size"];
                w = system_dict["dataset"]["params"]["input_size"];
            
            system_dict["local"]["transforms_test"].append(transforms.Resize(size=(w, h)));
            system_dict["local"]["transforms_test"].append(transforms.ToTensor());
            system_dict["local"]["applied_test_tensor"] = True;
    system_dict["local"]["data_transforms"]["test"] = transforms.Compose(system_dict["local"]["transforms_test"]);

    return system_dict;


import os
import sys

import psutil

from monk.keras_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_unit_tests.keras.common import print_start
from monk.pip_unit_tests.keras.common import print_status

import tensorflow as tf
if(tf.__version__[0] == '2'):
    import tensorflow.compat.v1 as tf
    tf.disable_v2_behavior()
import numpy as np


def test_layer_max_pooling2d(system_dict):
    forward = True;

    test = "test_layer_max_pooling2d";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf = prototype(verbose=0);
            gtf.Prototype("sample-project-1", "sample-experiment-1");


            network = [];
            network.append(gtf.max_pooling2d(kernel_size=3));
            gtf.Compile_Network(network, data_shape=(3, 32, 32), use_gpu=False);

            x = tf.placeholder(tf.float32, shape=(1, 32, 32, 3))
            y = gtf.system_dict["local"]["model"](x);          

            system_dict["successful_tests"] += 1;
            print_status("Pass");

        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    return system_dict



from monk.gluon.finetune.imports import *
from monk.system.imports import *
from monk.gluon.finetune.level_2_model_base import finetune_model



class finetune_training(finetune_model):
    '''
    Base class for training and associated functions

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def get_training_estimate(self):
        '''
        Get estimated time for training a single epoch based on all set parameters

        Args:
            None

        Returns:
            float: Total time per epoch in seconds
        '''
        total_time_per_epoch = 0;

        self.system_dict = load_scheduler(self.system_dict);
        self.system_dict = load_optimizer(self.system_dict);
        self.system_dict = load_loss(self.system_dict);


        num_iterations_train = len(self.system_dict["local"]["data_loaders"]["train"])//10;
        num_iterations_val = len(self.system_dict["local"]["data_loaders"]["val"])//10;


        since = time.time();
        train_loss = 0;
        for i, batch in enumerate(self.system_dict["local"]["data_loaders"]["train"]):
            data = mx.gluon.utils.split_and_load(batch[0], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
            label = mx.gluon.utils.split_and_load(batch[1], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
            with ag.record():
                outputs = [self.system_dict["local"]["model"](X) for X in data]
                loss = [self.system_dict["local"]["criterion"](yhat, y) for yhat, y in zip(outputs, label)]
            for l in loss:
                l.backward()

            train_loss += sum([l.mean().asscalar() for l in loss]) / len(loss)
            
            if(i==num_iterations_train):
                break;

        for i, batch in enumerate(self.system_dict["local"]["data_loaders"]["val"]):
            data = mx.gluon.utils.split_and_load(batch[0], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
            label = mx.gluon.utils.split_and_load(batch[1], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
            
            with ag.record():
                outputs = [self.system_dict["local"]["model"](X) for X in data]
                loss = [self.system_dict["local"]["criterion"](yhat, y) for yhat, y in zip(outputs, label)]
            
            
            if(i==num_iterations_val):
                break;

        total_time_per_epoch = (time.time() - since)*10;
        
        return total_time_per_epoch;
    ###############################################################################################################################################





    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_training_evaluation(self):
        '''
        Base function for running validation while training

        Args:
            None

        Returns:
            dict: Validation metrics
            float: Test Loss 
        '''
        num_batch = len(self.system_dict["local"]["data_loaders"]["val"]);

        if(self.system_dict["dataset"]["label_type"] == "single"):
            metric = mx.metric.Accuracy();
        else:
            metric = mx.metric.CustomMetric(feval=self.custom_metric)
            
        if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
            pbar = tqdm(total=num_batch);
        
        test_loss = 0;
        
        for i, batch in enumerate(self.system_dict["local"]["data_loaders"]["val"]):
            if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                pbar.update();

            data = mx.gluon.utils.split_and_load(batch[0], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
            label = mx.gluon.utils.split_and_load(batch[1], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
            
            with ag.record():
                outputs = [self.system_dict["local"]["model"](X) for X in data]
                loss = [self.system_dict["local"]["criterion"](yhat, y) for yhat, y in zip(outputs, label)]
            test_loss += sum([l.mean().asscalar() for l in loss]) / len(loss)
            metric.update(label, outputs)
        return metric.get(), test_loss;
    ###############################################################################################################################################




    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_training_final(self):
        '''
        Main training function

        Args:
            None

        Returns:
            None
        '''
        if(self.system_dict["states"]["resume_train"]):
            self.custom_print("Training Resume");
            self.system_dict = load_scheduler(self.system_dict);
            self.system_dict = load_optimizer(self.system_dict);
            self.system_dict = load_loss(self.system_dict);

            if(self.system_dict["dataset"]["label_type"] == "single"):
                metric = mx.metric.Accuracy();
            else:
                metric = mx.metric.CustomMetric(feval=self.custom_metric)

            trainer = mx.gluon.Trainer(self.system_dict["local"]["model"].collect_params(), optimizer=self.system_dict["local"]["optimizer"]);

            self.system_dict["training"]["status"] = False;
            since = time.time()
            pid = os.getpid();

            if(self.system_dict["training"]["settings"]["save_training_logs"]):
                val_acc_history = list(np.load(self.system_dict["log_dir"] + "val_acc_history.npy", allow_pickle=True));
                train_acc_history = list(np.load(self.system_dict["log_dir"] + "train_acc_history.npy", allow_pickle=True));
                val_loss_history = list(np.load(self.system_dict["log_dir"] + "val_loss_history.npy", allow_pickle=True));
                train_loss_history = list(np.load(self.system_dict["log_dir"] + "train_loss_history.npy", allow_pickle=True));

            num_batch_train = len(self.system_dict["local"]["data_loaders"]["train"]);
            num_batch_val = len(self.system_dict["local"]["data_loaders"]["val"]);

            best_acc = 0.0;
            best_acc_epoch = 0;
            max_gpu_usage = 0;

            for epoch in range(self.system_dict["hyper-parameters"]["num_epochs"]):
                if(self.system_dict["training"]["settings"]["display_progress"]):
                    self.custom_print('    Epoch {}/{}'.format(epoch+1, self.system_dict["hyper-parameters"]["num_epochs"]))
                    self.custom_print('    ' + '-' * 10)

                if(epoch < self.system_dict["training"]["outputs"]["epochs_completed"]):
                    self.custom_print("Skipping Current Epoch");
                    self.custom_print("");
                    self.custom_print("");
                    continue;

                since = time.time();
                train_loss = 0
                metric.reset()
                
                if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                    pbar = tqdm(total=num_batch_train);
                
                for i, batch in enumerate(self.system_dict["local"]["data_loaders"]["train"]):
                    if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                        pbar.update();

                    data = mx.gluon.utils.split_and_load(batch[0], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
                    label = mx.gluon.utils.split_and_load(batch[1], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
    
                    with ag.record():
                        outputs = [self.system_dict["local"]["model"](X) for X in data]
                        loss = [self.system_dict["local"]["criterion"](yhat, y) for yhat, y in zip(outputs, label)]
                    for l in loss:
                        l.backward()

                    trainer.step(self.system_dict["dataset"]["params"]["batch_size"]);
                    train_loss += sum([l.mean().asscalar() for l in loss]) / len(loss)
                    metric.update(label, outputs)

                _, train_acc = metric.get()
                train_loss /= num_batch_train;

                val_acc, val_loss = self.set_training_evaluation();
                val_acc = val_acc[1];
                val_loss /= num_batch_val;

                

                if(not os.getcwd() == "/kaggle/working"):
                    if(self.system_dict["model"]["params"]["use_gpu"]):
                        GPUs = GPUtil.getGPUs()
                        gpuMemoryUsed = GPUs[0].memoryUsed
                        if(self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] < int(gpuMemoryUsed)):
                            self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = int(gpuMemoryUsed);
                else:
                    gpuMemoryUsed = 0;
                    self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = 0;

                if(self.system_dict["training"]["settings"]["save_training_logs"]):
                    val_acc_history.append(val_acc);
                    val_loss_history.append(val_loss);
                    train_acc_history.append(train_acc);
                    train_loss_history.append(train_loss);

                if(val_acc > best_acc):
                    best_acc = val_acc;
                    best_acc_epoch = epoch;
                    if(self.system_dict["training"]["settings"]["save_intermediate_models"]):
                        self.system_dict["local"]["model"].export(self.system_dict["model_dir"] + self.system_dict["training"]["settings"]["intermediate_model_prefix"],
                            epoch=epoch)
                    self.system_dict["local"]["model"].export(self.system_dict["model_dir"] + "best_model", epoch=0);

                    self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
                    self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;

                time_elapsed_since = time.time() - since;
                if("training_time" in self.system_dict["training"]["outputs"].keys()):
                    minutes, seconds = self.system_dict["training"]["outputs"]["training_time"].split(" ");
                    minutes = int(minutes[:len(minutes)-1]);
                    seconds = int(seconds[:len(seconds)-1]);
                    time_elapsed_since += minutes*60 + seconds;
                self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);

                if(self.system_dict["training"]["settings"]["save_training_logs"]):
                    np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history));
                    np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history));
                    np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history));
                    np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history));

                    create_train_test_plots_accuracy([train_acc_history, val_acc_history], ["Epoch Num", "Accuracy"], self.system_dict["log_dir"], show_img=False, save_img=True);
                    create_train_test_plots_loss([train_loss_history, val_loss_history], ["Epoch Num", "Loss"], self.system_dict["log_dir"], show_img=False, save_img=True);
                
                self.system_dict["local"]["model"].export(self.system_dict["model_dir"] + "resume_state", epoch=0);
                if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                    self.custom_print("");
                    self.custom_print("");

                if(self.system_dict["training"]["settings"]["display_progress"]):
                    curr_lr = trainer.learning_rate
                    self.custom_print("    curr_lr - {}".format(curr_lr));
                    self.custom_print('    [Epoch %d] Train-acc: %.3f, Train-loss: %.3f | Val-acc: %3f, Val-loss: %.3f, | time: %.1f sec' %
                             (epoch+1, train_acc, train_loss, val_acc, val_loss, time.time() - since));
                    self.custom_print("");
                self.system_dict["training"]["outputs"]["epochs_completed"] = epoch+1;
                save(self.system_dict);    



        elif(self.system_dict["states"]["eval_infer"]):
            msg = "Cannot train in testing (eval_infer) mode.\n";
            msg += "Tip - use new_experiment function with a copy_from argument.\n";
            raise ConstraintError(msg);
       

        else:
            self.custom_print("Training Start");
            self.system_dict = load_scheduler(self.system_dict);
            self.system_dict = load_optimizer(self.system_dict);
            self.system_dict = load_loss(self.system_dict);

            if(self.system_dict["dataset"]["label_type"] == "single"):
                metric = mx.metric.Accuracy();
            else:
                metric = mx.metric.CustomMetric(feval=self.custom_metric)
            trainer = mx.gluon.Trainer(self.system_dict["local"]["model"].collect_params(), optimizer=self.system_dict["local"]["optimizer"]);

            self.system_dict["training"]["status"] = False;

            pid = os.getpid()

            if(self.system_dict["training"]["settings"]["save_training_logs"]):
                val_acc_history = [];
                train_acc_history = [];
                val_loss_history = [];
                train_loss_history = [];

            num_batch_train = len(self.system_dict["local"]["data_loaders"]["train"]);
            num_batch_val = len(self.system_dict["local"]["data_loaders"]["val"]);

            best_acc = 0.0;
            best_acc_epoch = 0;
            max_gpu_usage = 0;

            for epoch in range(self.system_dict["hyper-parameters"]["num_epochs"]):
                if(self.system_dict["training"]["settings"]["display_progress"]):
                    self.custom_print('    Epoch {}/{}'.format(epoch+1, self.system_dict["hyper-parameters"]["num_epochs"]))
                    self.custom_print('    ' + '-' * 10)

                since = time.time();
                train_loss = 0
                metric.reset()
                
                if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                    pbar = tqdm(total=num_batch_train);
                
                for i, batch in enumerate(self.system_dict["local"]["data_loaders"]["train"]):
                    if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                        pbar.update();

                    data = mx.gluon.utils.split_and_load(batch[0], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
                    label = mx.gluon.utils.split_and_load(batch[1], ctx_list=self.system_dict["local"]["ctx"], batch_axis=0, even_split=False)
    
                    with ag.record():
                        outputs = [self.system_dict["local"]["model"](X) for X in data]
                        loss = [self.system_dict["local"]["criterion"](yhat, y) for yhat, y in zip(outputs, label)]
                    for l in loss:
                        l.backward()

                    trainer.step(self.system_dict["dataset"]["params"]["batch_size"]);
                    train_loss += sum([l.mean().asscalar() for l in loss]) / len(loss)
                    metric.update(label, outputs)

                _, train_acc = metric.get()
                train_loss /= num_batch_train;

                val_acc, val_loss = self.set_training_evaluation();
                val_acc = val_acc[1];
                val_loss /= num_batch_val;

                

                if(not os.getcwd() == "/kaggle/working"):
                    if(self.system_dict["model"]["params"]["use_gpu"]):
                        GPUs = GPUtil.getGPUs()
                        gpuMemoryUsed = GPUs[0].memoryUsed
                        if(self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] < int(gpuMemoryUsed)):
                            self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = int(gpuMemoryUsed);
                else:
                    gpuMemoryUsed = 0;
                    self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = 0;

                if(self.system_dict["training"]["settings"]["save_training_logs"]):
                    val_acc_history.append(val_acc);
                    val_loss_history.append(val_loss);
                    train_acc_history.append(train_acc);
                    train_loss_history.append(train_loss);

                if(self.system_dict["training"]["settings"]["save_intermediate_models"]):
                    self.system_dict["local"]["model"].export(self.system_dict["model_dir"] + self.system_dict["training"]["settings"]["intermediate_model_prefix"],
                            epoch=epoch)


                if(val_acc > best_acc):
                    best_acc = val_acc;
                    best_acc_epoch = epoch;
                    self.system_dict["local"]["model"].export(self.system_dict["model_dir"] + "best_model", epoch=0);
                    self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
                    self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;

                time_elapsed_since = time.time() - since;
                if("training_time" in self.system_dict["training"]["outputs"].keys()):
                    minutes, seconds = self.system_dict["training"]["outputs"]["training_time"].split(" ");
                    minutes = int(minutes[:len(minutes)-1]);
                    seconds = int(seconds[:len(seconds)-1]);
                    time_elapsed_since += minutes*60 + seconds;
                self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);

                if(self.system_dict["training"]["settings"]["save_training_logs"]):
                    np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history));
                    np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history));
                    np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history));
                    np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history));

                    create_train_test_plots_accuracy([train_acc_history, val_acc_history], ["Epoch Num", "Accuracy"], self.system_dict["log_dir"], show_img=False, save_img=True);
                    create_train_test_plots_loss([train_loss_history, val_loss_history], ["Epoch Num", "Loss"], self.system_dict["log_dir"], show_img=False, save_img=True);
                
                self.system_dict["local"]["model"].export(self.system_dict["model_dir"] + "resume_state", epoch=0);
                if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                    self.custom_print("");
                    self.custom_print("");

                if(self.system_dict["training"]["settings"]["display_progress"]):
                    curr_lr = trainer.learning_rate
                    self.custom_print("    curr_lr - {}".format(curr_lr));
                    self.custom_print('    [Epoch %d] Train-acc: %.3f, Train-loss: %.3f | Val-acc: %3f, Val-loss: %.3f, | time: %.1f sec' %
                             (epoch+1, train_acc, train_loss, val_acc, val_loss, time.time() - since));
                    self.custom_print("");
                self.system_dict["training"]["outputs"]["epochs_completed"] = epoch+1;
                save(self.system_dict);

            if(self.system_dict["training"]["settings"]["display_progress"]):
                self.custom_print('    Training completed in: {:.0f}m {:.0f}s'.format(time_elapsed_since // 60, time_elapsed_since % 60))
                self.custom_print('    Best val Acc:          {:4f}'.format(best_acc))
                self.custom_print("");


        if(not self.system_dict["states"]["eval_infer"]):
            self.custom_print("Training End");
            self.custom_print("");
            self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
            self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;
            self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);
            self.system_dict["training"]["outputs"]["max_gpu_usage"] = str(self.system_dict["training"]["outputs"]["max_gpu_memory_usage"]) + " Mb";

            self.system_dict["local"]["model"].export(self.system_dict["model_dir"] + "final", epoch=0);

            if(self.system_dict["training"]["settings"]["save_training_logs"]): 
                self.custom_print("Training Outputs");
                self.custom_print("    Model Dir:   {}".format(self.system_dict["model_dir"]));
                self.custom_print("    Log Dir:     {}".format(self.system_dict["log_dir"]));
                self.custom_print("    Final model: {}".format("final"));
                self.custom_print("    Best model:  {}".format("best_model"));
                self.custom_print("    Log 1 - Validation accuracy history log: {}".format("val_acc_history.npy"));
                self.custom_print("    Log 2 - Validation loss history log:     {}".format("val_loss_history.npy"));
                self.custom_print("    Log 3 - Training accuracy history log:   {}".format("train_acc_history.npy"));
                self.custom_print("    Log 4 - Training loss history log:       {}".format("train_loss_history.npy"));
                self.custom_print("    Log 5 - Training curve:                  {}".format("train_loss_history.npy"));
                self.custom_print("    Log 6 - Validation curve:                {}".format("train_loss_history.npy"));
                self.custom_print("");

                np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history));
                np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history));
                np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history));
                np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history));
                
                self.system_dict["training"]["outputs"]["log_val_acc_history"] = self.system_dict["log_dir"] + "val_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_val_loss_history"] = self.system_dict["log_dir"] + "val_loss_history.npy";
                self.system_dict["training"]["outputs"]["log_train_acc_history"] = self.system_dict["log_dir"] + "train_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_train_loss_history"] = self.system_dict["log_dir"] + "train_loss_history.npy";

                self.system_dict["training"]["outputs"]["log_val_acc_history_relative"] = self.system_dict["log_dir_relative"] + "val_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_val_loss_history_relative"] = self.system_dict["log_dir_relative"] + "val_loss_history.npy";
                self.system_dict["training"]["outputs"]["log_train_acc_history_relative"] = self.system_dict["log_dir_relative"] + "train_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_train_loss_history_relative"] = self.system_dict["log_dir_relative"] + "train_loss_history.npy";


                create_train_test_plots_accuracy([train_acc_history, val_acc_history], ["Epoch Num", "Accuracy"], self.system_dict["log_dir"], show_img=False, save_img=True);
                create_train_test_plots_loss([train_loss_history, val_loss_history], ["Epoch Num", "Loss"], self.system_dict["log_dir"], show_img=False, save_img=True);

            self.system_dict["training"]["status"] = True;
        

    ###############################################################################################################################################






    ###############################################################################################################################################
    def custom_metric(self, labels, raw_scores):
        num_correct = 0;
        total_labels = 0;

        list_classes = [];

        for i in range(labels.shape[0]):
            for j in range(labels.shape[1]):
                score = logistic.cdf(raw_scores[i][j])
                pred = False
                if(score > 0.5):
                    pred = True
                else:
                    pred = False
                if(pred and labels[i][j]):
                    num_correct += 1;

                if(labels[i][j]):
                    total_labels += 1;

        return num_correct/total_labels;

    ###############################################################################################################################################





from monk.tf_keras_1.models.imports import *
from monk.system.imports import *
from monk.tf_keras_1.models.common import set_parameter_requires_grad


#classifier 6
set1 = ["mobilenet", "densenet121", "densenet169", "densenet201", "inception_v3", 
        "inception_resnet_v3", "mobilenet_v2", "nasnet_mobile", "nasnet_large", "resnet50",
        "resnet101", "resnet152", "resnet50_v2", "resnet101_v2", "resnet152_v2", "vgg16",
        "vgg19", "xception"];

combined_list = set1
combined_list_lower = list(map(str.lower, combined_list))


@accepts(str, bool, int, bool, int, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def get_base_model(model_name, use_pretrained, num_classes, freeze_base_network, input_size):
    '''
    Get base network for transfer learning based on parameters selected

    Args:
        model_name (str): Select from available models. Check via List_Models() function
        freeze_base_network (bool): If set as True, then base network's weights are freezed (cannot be trained)
        use_gpu (bool): If set as True, uses GPU
        use_pretrained (bool): If set as True, use weights trained on imagenet and coco like dataset
                                Else, use randomly initialized weights.

    Returns:
        neural network: Base network
        str: Name of the model
    '''
    if(model_name not in combined_list_lower):
        print("Model name: {} not found".format(model_name));
    else:
        index = combined_list_lower.index(model_name);
        model_name = combined_list[index];

        if(use_pretrained):
            weights="imagenet";
        else:
            weights=None;


        if(model_name == "mobilenet"):
            from keras.applications import MobileNet as keras_model
        elif(model_name == "densenet121"):
            from keras.applications import DenseNet121 as keras_model
        elif(model_name == "densenet169"):
            from keras.applications import DenseNet169 as keras_model
        elif(model_name == "densenet201"):
            from keras.applications import DenseNet201 as keras_model
        elif(model_name == "inception_v3"):
            from keras.applications import InceptionV3 as keras_model
        elif(model_name == "inception_resnet_v3"):
            from keras.applications import InceptionResNetV2 as keras_model
        elif(model_name == "mobilenet_v2"):
            from keras.applications import MobileNetV2 as keras_model
        elif(model_name == "nasnet_mobile"):
            from keras.applications import NASNetMobile as keras_model
        elif(model_name == "nasnet_large"):
            from keras.applications import NASNetLarge as keras_model
        elif(model_name == "resnet50"):
            from keras.applications import ResNet50 as keras_model
        elif(model_name == "resnet101"):
            from keras.applications import ResNet101 as keras_model
        elif(model_name == "resnet152"):
            from keras.applications import ResNet152 as keras_model
        elif(model_name == "resnet50_v2"):
            from keras.applications import ResNet50V2 as keras_model
        elif(model_name == "resnet101_v2"):
            from keras.applications import ResNet101V2 as keras_model
        elif(model_name == "resnet152_v2"):
            from keras.applications import ResNet152V2 as keras_model
        elif(model_name == "vgg16"):
            from keras.applications import VGG16 as keras_model
        elif(model_name == "vgg19"):
            from keras.applications import VGG19 as keras_model
        elif(model_name == "xception"):
            from keras.applications import Xception as keras_model


        finetune_net = keras_model(weights=weights, include_top=False, input_shape=(input_size, input_size, 3));

        finetune_net = set_parameter_requires_grad(finetune_net, freeze_base_network);

        return finetune_net, model_name;


from monk.pytorch.optimizers.imports import *
from monk.system.imports import *

@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def load_optimizer(system_dict):
    '''
    Load Optimizers in training states

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    optimizer = system_dict["local"]["optimizer"];
    learning_rate = system_dict["hyper-parameters"]["learning_rate"];

    
    if(optimizer == "sgd"):
        system_dict["local"]["optimizer"] = torch.optim.SGD(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            momentum=system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"], 
            dampening=system_dict["hyper-parameters"]["optimizer"]["params"]["momentum_dampening_rate"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
            nesterov=False);

    elif(optimizer == "nesterov_sgd"):
        system_dict["local"]["optimizer"] = torch.optim.SGD(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            momentum=system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"], 
            dampening=system_dict["hyper-parameters"]["optimizer"]["params"]["momentum_dampening_rate"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
            nesterov=False);

    elif(optimizer == "rmsprop"):
        system_dict["local"]["optimizer"] = torch.optim.RMSprop(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            alpha=system_dict["hyper-parameters"]["optimizer"]["params"]["decay_rate"], 
            eps=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
            momentum=0.0, 
            centered=False);

    elif(optimizer == "momentum_rmsprop"):
        system_dict["local"]["optimizer"] = torch.optim.RMSprop(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            alpha=system_dict["hyper-parameters"]["optimizer"]["params"]["decay_rate"], 
            eps=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
            momentum=system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"], 
            centered=True);


    elif(optimizer == "adam"):
        system_dict["local"]["optimizer"] = torch.optim.Adam(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            betas=(system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"], system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"]),  
            eps=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
            amsgrad=system_dict["hyper-parameters"]["optimizer"]["params"]["amsgrad"]);


    elif(optimizer == "adamax"):
        system_dict["local"]["optimizer"] = torch.optim.Adamax(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            betas=(system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"], system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"]), 
            eps=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"]);


    elif(optimizer == "adamw"):
        system_dict["local"]["optimizer"] = torch.optim.AdamW(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            betas=(system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"], system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"]),  
            eps=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
            amsgrad=system_dict["hyper-parameters"]["optimizer"]["params"]["amsgrad"]);


    elif(optimizer == "adadelta"):
        system_dict["local"]["optimizer"] = torch.optim.Adadelta(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            rho=system_dict["hyper-parameters"]["optimizer"]["params"]["rho"], 
            eps=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"]);

    elif(optimizer == "adagrad"):
        system_dict["local"]["optimizer"] = torch.optim.Adagrad(
            system_dict["local"]["params_to_update"], 
            lr=learning_rate, 
            lr_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["lr_decay"], 
            weight_decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
            initial_accumulator_value=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"]);


    return system_dict;


from monk.pytorch.datasets.imports import *
from monk.system.imports import *


@accepts([int, tuple], dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_input_size(input_size, system_dict):
    '''
    Set Input data size

    Args:
        input_size (int, tuple): Single integer in case of square shaped image
                                  Tuple representing (width, height)
        system_dict (dict): System Dictionary
    Returns:
        dict: Updated system dictionary.
    '''
    system_dict["dataset"]["params"]["input_size"] = input_size;
    return system_dict;


@accepts(int, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_batch_size(batch_size, system_dict):
    '''
    Set Input batch size

    Args:
        batch_size (int): Batch sizes for sampling input data during training
        system_dict (dict): System Dictionary
    Returns:
        dict: Updated system dictionary.
    '''
    system_dict["dataset"]["params"]["batch_size"] = batch_size;
    return system_dict;
    
@accepts(bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_data_shuffle(value, system_dict):
    '''
    Set weather to shuffle data

    Args:
        value (bool): If True data is shuffled before sampling into batches
        system_dict (dict): System Dictionary
    Returns:
        dict: Updated system dictionary.
    '''
    train_shuffle = value;
    val_shuffle = value;
    system_dict["dataset"]["params"]["train_shuffle"] = train_shuffle;
    system_dict["dataset"]["params"]["val_shuffle"] = val_shuffle;
    return system_dict;
    

@accepts(int, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_num_processors(num_workers, system_dict):
    '''
    Set Number of processors for data sampling

    Args:
        num_workers (int): Max number of CPUs to be used
        system_dict (dict): System Dictionary
    Returns:
        dict: Updated system dictionary.
    '''
    system_dict["dataset"]["params"]["num_workers"] = num_workers;
    return system_dict;


@accepts(bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_weighted_sampling(sample, system_dict):
    '''
            
    '''
    system_dict["dataset"]["params"]["weighted_sample"] = sample;
    return system_dict;


import os
import sys

import psutil

from monk.gluon_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_unit_tests.gluon.common import print_start
from monk.pip_unit_tests.gluon.common import print_status

import mxnet as mx
import numpy as np
from monk.gluon.losses.return_loss import load_loss


def test_block_inception_e(system_dict):
    forward = True;

    test = "test_block_inception_e";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf = prototype(verbose=0);
            gtf.Prototype("sample-project-1", "sample-experiment-1");


            network = [];
            network.append(gtf.inception_e_block(pool_type="avg"));
            network.append(gtf.inception_e_block(pool_type="max"));
            gtf.Compile_Network(network, use_gpu=False);

            x = np.random.rand(1, 1, 64, 64);
            x = mx.nd.array(x);
            y = gtf.system_dict["local"]["model"].forward(x);          

            system_dict["successful_tests"] += 1;
            print_status("Pass");

        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    return system_dict



from monk.tf_keras_1.finetune.imports import *
from monk.system.imports import *
from monk.tf_keras_1.finetune.level_2_model_base import finetune_model



class finetune_training(finetune_model):
    '''
    Base class for training and associated functions

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);



    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def get_training_estimate(self):
        '''
        Get estimated time for training a single epoch based on all set parameters

        Args:
            None

        Returns:
            float: Total time per epoch in seconds
        '''
        total_time_per_epoch = 0;


        self.system_dict = load_scheduler(self.system_dict);
        self.system_dict = load_optimizer(self.system_dict);
        self.system_dict = load_loss(self.system_dict);

        self.system_dict["local"]["model"].compile(optimizer=self.system_dict["local"]["optimizer"], 
                loss=self.system_dict["local"]["criterion"], metrics=['accuracy']);

        time_callback = TimeHistory();
        initial_epoch = 0;

        step_size_estimate = self.system_dict["local"]["data_loaders"]["estimate"].n//self.system_dict["local"]["data_loaders"]["estimate"].batch_size;

        self.system_dict["local"]["model"].fit_generator(generator=self.system_dict["local"]["data_loaders"]["estimate"],
                           steps_per_epoch=step_size_estimate,
                           epochs=1,
                           callbacks=[time_callback],
                           workers=psutil.cpu_count(),
                           initial_epoch = initial_epoch,
                           verbose=0)

        time_taken = time_callback.times[0];
        num_images = len(self.system_dict["local"]["data_loaders"]["estimate"].labels);
        time_taken_per_image = time_taken/num_images;

        total_time_per_epoch = time_taken_per_image*(self.system_dict["dataset"]["params"]["num_train_images"] + 
            self.system_dict["dataset"]["params"]["num_val_images"]);

        return total_time_per_epoch;

    ###############################################################################################################################################
        


    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_training_final(self):
        '''
        Main training function

        Args:
            None

        Returns:
            None
        '''
        if(self.system_dict["states"]["resume_train"]):
            self.custom_print("Training Resume");
            total_time_per_epoch = 0;

            self.system_dict = load_scheduler(self.system_dict);
            self.system_dict = load_optimizer(self.system_dict);
            self.system_dict = load_loss(self.system_dict);

            self.system_dict["training"]["status"] = False;

            pid = os.getpid();


            log_dir = self.system_dict["log_dir_relative"];
            model_dir = self.system_dict["model_dir_relative"];
            intermediate_model_prefix = self.system_dict["training"]["settings"]["intermediate_model_prefix"];
            save_intermediate_models = self.system_dict["training"]["settings"]["save_intermediate_models"];
            display_progress_realtime = self.system_dict["training"]["settings"]["display_progress_realtime"];
            display_progress = self.system_dict["training"]["settings"]["display_progress"];
            num_epochs = self.system_dict["hyper-parameters"]["num_epochs"];

            f = open(self.system_dict["log_dir"] + "/model_history_log.csv", 'r');
            lines = f.readlines();
            f.close();
            epochs_completed = len(lines)-1;

            if(self.system_dict["training"]["settings"]["save_training_logs"]):
                history_df = pd.read_csv(self.system_dict["log_dir"] + "/model_history_log.csv");
                if(int(keras.__version__.split(".")[1]) > 2):
                    val_acc_history = history_df['val_accuracy'].tolist();
                    train_acc_history = history_df['accuracy'].tolist();
                    val_loss_history = history_df['val_loss'].tolist();
                    train_loss_history = history_df['loss'].tolist();
                else:
                    val_acc_history = history_df['val_acc'].tolist();
                    train_acc_history = history_df['acc'].tolist();
                    val_loss_history = history_df['val_loss'].tolist();
                    train_loss_history = history_df['loss'].tolist();

            f = open(self.system_dict["log_dir"] + "/times.txt", 'r');
            lines = f.readlines();
            times_history = [];
            for i in range(len(lines)):
                times_history.append(float(lines[i][:len(lines[i])-1]));


            csv_logger = krc.CSVLogger(log_dir + "model_history_log.csv", append=True);

            if(not self.system_dict["verbose"]):
                verbose=0;
            elif(display_progress_realtime):
                verbose=1;
            elif(display_progress):
                verbose=2;
            else:
                verbose=0;
            

            ckpt_all = krc.ModelCheckpoint(model_dir + intermediate_model_prefix + '{epoch:02d}.h5', monitor='val_loss', verbose=verbose, 
                save_best_only=False, save_weights_only=False, mode='auto', period=1);

            ckpt_best = krc.ModelCheckpoint(model_dir + 'best_model.h5', monitor='val_loss', verbose=verbose, 
                save_best_only=True, save_weights_only=False, mode='auto', period=1);

            resume = krc.ModelCheckpoint(model_dir + 'resume_state.h5', monitor='val_loss', verbose=verbose, 
                save_best_only=False, save_weights_only=False, mode='auto', period=1);
            
            time_callback = TimeHistory(log_dir)
            memory_callback = MemoryHistory();


            callbacks = [krc.History(), memory_callback, time_callback, csv_logger, resume, ckpt_best];

            if(save_intermediate_models):
                callbacks.append(ckpt_all);

            if(self.system_dict["local"]["learning_rate_scheduler"]):
                if(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] == "reduceonplateaulr"):
                    callbacks.append(self.system_dict["local"]["learning_rate_scheduler"]);
                else:
                    callbacks.append(krc.LearningRateScheduler(self.system_dict["local"]["learning_rate_scheduler"]));


            step_size_train = self.system_dict["local"]["data_loaders"]["train"].n//self.system_dict["local"]["data_loaders"]["train"].batch_size;
            step_size_val = self.system_dict["local"]["data_loaders"]["val"].n//self.system_dict["local"]["data_loaders"]["val"].batch_size;


            initial_epoch = epochs_completed;

            


            self.system_dict["local"]["model"].compile(optimizer=self.system_dict["local"]["optimizer"], 
                loss=self.system_dict["local"]["criterion"], metrics=['accuracy']);


            history = self.system_dict["local"]["model"].fit_generator(generator=self.system_dict["local"]["data_loaders"]["train"],
                               steps_per_epoch=step_size_train,
                               epochs=num_epochs,
                               callbacks=callbacks,
                               validation_data=self.system_dict["local"]["data_loaders"]["val"],
                               validation_steps=step_size_val,
                               initial_epoch = initial_epoch,
                               verbose=verbose);

            time_elapsed_since = 0;
            times_history += time_callback.times
            for i in range(len(times_history)):
                time_elapsed_since += times_history[i];


            self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);

            if(int(keras.__version__.split(".")[1]) > 2):
                val_acc_history = history.history['val_accuracy'];
                val_loss_history = history.history['val_loss'];
                train_acc_history = history.history['accuracy'];
                train_loss_history = history.history['loss'];
            else:
                val_acc_history = history.history['val_acc'];
                val_loss_history = history.history['val_loss'];
                train_acc_history = history.history['acc'];
                train_loss_history = history.history['loss'];


            self.system_dict["training"]["outputs"]["best_val_acc"] = max(val_acc_history);
            self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = val_acc_history.index(self.system_dict["training"]["outputs"]["best_val_acc"]);
            self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = memory_callback.max_gpu_usage;


            if(self.system_dict["training"]["settings"]["save_training_logs"]):
                np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history));
                np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history));
                np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history));
                np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history));

                

            if(self.system_dict["training"]["settings"]["display_progress"]):
                self.custom_print('    Training completed in: {:.0f}m {:.0f}s'.format(time_elapsed_since // 60, time_elapsed_since % 60))
                self.custom_print('    Best val Acc:          {:4f}'.format(self.system_dict["training"]["outputs"]["best_val_acc"]))
                self.custom_print("");



        elif(self.system_dict["states"]["eval_infer"]):
            msg = "Cannot train in testing (eval_infer) mode.\n";
            msg += "Tip - use new_experiment function with a copy_from argument.\n";
            raise ConstraintError(msg);



        else:
            self.custom_print("Training Start");

            self.system_dict = load_scheduler(self.system_dict);
            self.system_dict = load_optimizer(self.system_dict);
            self.system_dict = load_loss(self.system_dict);

            self.system_dict["training"]["status"] = False;

            pid = os.getpid();


            log_dir = self.system_dict["log_dir_relative"];
            model_dir = self.system_dict["model_dir_relative"];
            intermediate_model_prefix = self.system_dict["training"]["settings"]["intermediate_model_prefix"];
            save_intermediate_models = self.system_dict["training"]["settings"]["save_intermediate_models"];
            display_progress_realtime = self.system_dict["training"]["settings"]["display_progress_realtime"];
            display_progress = self.system_dict["training"]["settings"]["display_progress"];
            num_epochs = self.system_dict["hyper-parameters"]["num_epochs"];


            if(not self.system_dict["verbose"]):
                verbose=0;
            elif(display_progress_realtime):
                verbose=1;
            elif(display_progress):
                verbose=2;
            else:
                verbose=0;


            csv_logger = krc.CSVLogger(log_dir + "model_history_log.csv", append=False);
            ckpt_all = krc.ModelCheckpoint(model_dir + intermediate_model_prefix + '{epoch:02d}.h5', monitor='val_loss', verbose=verbose, 
                save_best_only=False, save_weights_only=False, mode='auto', period=1);

            ckpt_best = krc.ModelCheckpoint(model_dir + 'best_model.h5', monitor='val_loss', verbose=verbose, 
                save_best_only=True, save_weights_only=False, mode='auto', period=1);

            resume = krc.ModelCheckpoint(model_dir + 'resume_state.h5', monitor='val_loss', verbose=verbose, 
                save_best_only=False, save_weights_only=False, mode='auto', period=1);
            
            time_callback = TimeHistory(log_dir)
            memory_callback = MemoryHistory();


            callbacks = [krc.History(), memory_callback, time_callback, csv_logger, resume, ckpt_best];

            if(save_intermediate_models):
                callbacks.append(ckpt_all);

            if(self.system_dict["local"]["learning_rate_scheduler"]):
                if(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] == "reduceonplateaulr"):
                    callbacks.append(self.system_dict["local"]["learning_rate_scheduler"]);
                else:
                    callbacks.append(krc.LearningRateScheduler(self.system_dict["local"]["learning_rate_scheduler"]));


            step_size_train = self.system_dict["local"]["data_loaders"]["train"].n//self.system_dict["local"]["data_loaders"]["train"].batch_size;
            step_size_val = self.system_dict["local"]["data_loaders"]["val"].n//self.system_dict["local"]["data_loaders"]["val"].batch_size;


            initial_epoch = 0;

            if(not self.system_dict["verbose"]):
                verbose=0;
            elif(display_progress_realtime):
                verbose=1;
            elif(display_progress):
                verbose=2;
            else:
                verbose=0;


            self.system_dict["local"]["model"].compile(optimizer=self.system_dict["local"]["optimizer"], 
                loss=self.system_dict["local"]["criterion"], metrics=['accuracy']);


            history = self.system_dict["local"]["model"].fit_generator(generator=self.system_dict["local"]["data_loaders"]["train"],
                               steps_per_epoch=step_size_train,
                               epochs=num_epochs,
                               callbacks=callbacks,
                               validation_data=self.system_dict["local"]["data_loaders"]["val"],
                               validation_steps=step_size_val,
                               initial_epoch = initial_epoch,
                               verbose=verbose);



            time_elapsed_since = 0;
            for i in range(len(time_callback.times)):
                time_elapsed_since += time_callback.times[i];


            self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);

            if(int(keras.__version__.split(".")[1]) > 2):
                val_acc_history = history.history['val_accuracy'];
                val_loss_history = history.history['val_loss'];
                train_acc_history = history.history['accuracy'];
                train_loss_history = history.history['loss'];
            else:
                val_acc_history = history.history['val_acc'];
                val_loss_history = history.history['val_loss'];
                train_acc_history = history.history['acc'];
                train_loss_history = history.history['loss'];

            self.system_dict["training"]["outputs"]["best_val_acc"] = max(val_acc_history);
            self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = val_acc_history.index(self.system_dict["training"]["outputs"]["best_val_acc"]);
            self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = memory_callback.max_gpu_usage;


            if(self.system_dict["training"]["settings"]["save_training_logs"]):
                np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history));
                np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history));
                np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history));
                np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history));

               


            if(self.system_dict["training"]["settings"]["display_progress"]):
                self.custom_print('    Training completed in: {:.0f}m {:.0f}s'.format(time_elapsed_since // 60, time_elapsed_since % 60))
                self.custom_print('    Best val Acc:          {:4f}'.format(self.system_dict["training"]["outputs"]["best_val_acc"]))
                self.custom_print("");

            

        self.system_dict["local"]["model"].save(self.system_dict["model_dir_relative"] + "final.h5");
        self.system_dict["training"]["status"] = True;
        save(self.system_dict);



        if(not self.system_dict["states"]["eval_infer"]):
            self.custom_print("Training End");
            self.custom_print("");
            self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(self.system_dict["training"]["outputs"]["best_val_acc"]);
            self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"];
            self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);
            self.system_dict["training"]["outputs"]["max_gpu_usage"] = str(self.system_dict["training"]["outputs"]["max_gpu_memory_usage"]) + " Mb";


            if(self.system_dict["training"]["settings"]["save_training_logs"]): 
                self.custom_print("Training Outputs");
                self.custom_print("    Model Dir:   {}".format(self.system_dict["model_dir"]));
                self.custom_print("    Log Dir:     {}".format(self.system_dict["log_dir"]));
                self.custom_print("    Final model: {}".format("final"));
                self.custom_print("    Best model:  {}".format("best_model"));
                self.custom_print("    Log 1 - Validation accuracy history log: {}".format("val_acc_history.npy"));
                self.custom_print("    Log 2 - Validation loss history log:     {}".format("val_loss_history.npy"));
                self.custom_print("    Log 3 - Training accuracy history log:   {}".format("train_acc_history.npy"));
                self.custom_print("    Log 4 - Training loss history log:       {}".format("train_loss_history.npy"));
                self.custom_print("    Log 5 - Training curve:                  {}".format("train_loss_history.npy"));
                self.custom_print("    Log 6 - Validation curve:                {}".format("train_loss_history.npy"));
                self.custom_print("");

                np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history));
                np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history));
                np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history));
                np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history));
                
                self.system_dict["training"]["outputs"]["log_val_acc_history"] = self.system_dict["log_dir"] + "val_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_val_loss_history"] = self.system_dict["log_dir"] + "val_loss_history.npy";
                self.system_dict["training"]["outputs"]["log_train_acc_history"] = self.system_dict["log_dir"] + "train_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_train_loss_history"] = self.system_dict["log_dir"] + "train_loss_history.npy";

                self.system_dict["training"]["outputs"]["log_val_acc_history_relative"] = self.system_dict["log_dir_relative"] + "val_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_val_loss_history_relative"] = self.system_dict["log_dir_relative"] + "val_loss_history.npy";
                self.system_dict["training"]["outputs"]["log_train_acc_history_relative"] = self.system_dict["log_dir_relative"] + "train_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_train_loss_history_relative"] = self.system_dict["log_dir_relative"] + "train_loss_history.npy";


                create_train_test_plots_accuracy([train_acc_history, val_acc_history], ["Epoch Num", "Accuracy"], self.system_dict["log_dir"], show_img=False, save_img=True);
                create_train_test_plots_loss([train_loss_history, val_loss_history], ["Epoch Num", "Loss"], self.system_dict["log_dir"], show_img=False, save_img=True);

            self.system_dict["training"]["status"] = True;

        ###############################################################################################################################################



from monk.system.imports import *
from monk.tf_keras_1.finetune.imports import *

from monk.system.base_class import system


class finetune_dataset(system):
    '''
    Base class for dataset params

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''

    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @accepts("self", test=bool, estimate=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_dataset_dataloader(self, test=False, estimate=False):
        '''
        Setup the dataloader.

        Args:
            test (bool): If True then test data is loaded, else train data is loaded.
            estimate (bool): When estimating training time, estimate is True

        Returns:
            None
        '''
        if(test):
            num_workers = self.system_dict["dataset"]["params"]["num_workers"];
            csv_test = self.system_dict["dataset"]["csv_test"];
            dataset_test_path = self.system_dict["dataset"]["test_path"];
            delimiter = self.system_dict["dataset"]["params"]["test_delimiter"];
            label_type = self.system_dict["dataset"]["label_type"];

            if(type(self.system_dict["dataset"]["params"]["input_size"]) == tuple or type(self.system_dict["dataset"]["params"]["input_size"]) == list):
                h = self.system_dict["dataset"]["params"]["input_size"][0];
                w = self.system_dict["dataset"]["params"]["input_size"][1];
            else:
                h = self.system_dict["dataset"]["params"]["input_size"];
                w = self.system_dict["dataset"]["params"]["input_size"];

            target_size = (h, w);
            color_mode='rgb';
            class_mode='categorical';

            if(label_type == "single" or label_type == False):
                if(self.system_dict["dataset"]["params"]["dataset_test_type"] == "foldered"):
                    self.system_dict["local"]["data_loaders"]["test"] = self.system_dict["local"]["data_generators"]["test"].flow_from_directory(dataset_test_path, 
                                                                     target_size=target_size,
                                                                     color_mode=color_mode,
                                                                     batch_size=1,
                                                                     class_mode=class_mode,
                                                                     shuffle=False
                                                                     );

                elif(self.system_dict["dataset"]["params"]["dataset_test_type"] == "csv"):
                    test_df, columns = parse_csv2(csv_test, delimiter);
                    self.system_dict["local"]["data_loaders"]["test"] = self.system_dict["local"]["data_generators"]["test"].flow_from_dataframe(
                                                                    dataframe=test_df,
                                                                    directory=dataset_test_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=1,
                                                                    class_mode=class_mode,
                                                                    shuffle=False
                                                                    );
            else:
                test_df, columns = parse_csv2_updated(csv_train, delimiter);
                self.system_dict["local"]["data_loaders"]["test"] = self.system_dict["local"]["data_generators"]["test"].flow_from_dataframe(
                                                                    dataframe=test_df,
                                                                    directory=dataset_test_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=1,
                                                                    class_mode=class_mode,
                                                                    shuffle=False
                                                                    );




            if(not self.system_dict["dataset"]["params"]["classes"]):
                self.system_dict["dataset"]["params"]["classes"] = self.system_dict["local"]["data_loaders"]["test"].class_indices;

            self.system_dict["dataset"]["params"]["num_test_images"] = len(self.system_dict["local"]["data_loaders"]["test"].labels);


        elif(estimate):
            sampled_dataset = None;
            image_datasets = {};
            dataset_type = self.system_dict["dataset"]["dataset_type"];
            dataset_train_path = self.system_dict["dataset"]["train_path"];
            dataset_val_path = self.system_dict["dataset"]["val_path"];
            csv_train = self.system_dict["dataset"]["csv_train"];
            csv_val = self.system_dict["dataset"]["csv_val"];
            train_val_split = self.system_dict["dataset"]["params"]["train_val_split"];
            delimiter = self.system_dict["dataset"]["params"]["delimiter"];
            batch_size = self.system_dict["dataset"]["params"]["batch_size"];
            shuffle = self.system_dict["dataset"]["params"]["train_shuffle"];
            num_workers = self.system_dict["dataset"]["params"]["num_workers"];
            label_type = self.system_dict["dataset"]["label_type"];

            if(type(self.system_dict["dataset"]["params"]["input_size"]) == tuple or type(self.system_dict["dataset"]["params"]["input_size"]) == list):
                h = self.system_dict["dataset"]["params"]["input_size"][0];
                w = self.system_dict["dataset"]["params"]["input_size"][1];
            else:
                h = self.system_dict["dataset"]["params"]["input_size"];
                w = self.system_dict["dataset"]["params"]["input_size"];

            target_size = (h, w);

            color_mode='rgb';
            class_mode='categorical';


            if(label_type == "single" or label_type == False):
                if("csv" in dataset_type):
                    train_df, columns = parse_csv2(csv_train, delimiter);
                    self.system_dict["local"]["data_loaders"]["estimate"] = self.system_dict["local"]["data_generators"]["estimate"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    subset="training",
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );
                else:
                    self.system_dict["local"]["data_loaders"]["estimate"] = self.system_dict["local"]["data_generators"]["estimate"].flow_from_directory(dataset_train_path, 
                                                                     target_size=target_size,
                                                                     color_mode=color_mode,
                                                                     batch_size=batch_size,
                                                                     class_mode=class_mode,
                                                                     shuffle=shuffle,
                                                                     subset='training');
            else:
                train_df, columns = parse_csv2_updated(csv_train, delimiter);
                self.system_dict["local"]["data_loaders"]["estimate"] = self.system_dict["local"]["data_generators"]["estimate"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    subset="training",
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );


        else:
            sampled_dataset = None;
            image_datasets = {};
            dataset_type = self.system_dict["dataset"]["dataset_type"];
            dataset_train_path = self.system_dict["dataset"]["train_path"];
            dataset_val_path = self.system_dict["dataset"]["val_path"];
            csv_train = self.system_dict["dataset"]["csv_train"];
            csv_val = self.system_dict["dataset"]["csv_val"];
            train_val_split = self.system_dict["dataset"]["params"]["train_val_split"];
            delimiter = self.system_dict["dataset"]["params"]["delimiter"];
            batch_size = self.system_dict["dataset"]["params"]["batch_size"];
            shuffle = self.system_dict["dataset"]["params"]["train_shuffle"];
            num_workers = self.system_dict["dataset"]["params"]["num_workers"];
            label_type = self.system_dict["dataset"]["label_type"];


            if(type(self.system_dict["dataset"]["params"]["input_size"]) == tuple or type(self.system_dict["dataset"]["params"]["input_size"]) == list):
                h = self.system_dict["dataset"]["params"]["input_size"][0];
                w = self.system_dict["dataset"]["params"]["input_size"][1];
            else:
                h = self.system_dict["dataset"]["params"]["input_size"];
                w = self.system_dict["dataset"]["params"]["input_size"];

            target_size = (h, w);
            
            color_mode='rgb';
            class_mode='categorical';



            if(label_type == "single" or label_type == False):
                if(dataset_type == "train"):
                    self.system_dict["local"]["data_loaders"]["train"] = self.system_dict["local"]["data_generators"]["train"].flow_from_directory(dataset_train_path, 
                                                                     target_size=target_size,
                                                                     color_mode=color_mode,
                                                                     batch_size=batch_size,
                                                                     class_mode=class_mode,
                                                                     shuffle=shuffle,
                                                                     subset='training');

                    self.system_dict["local"]["data_loaders"]["val"] = self.system_dict["local"]["data_generators"]["train"].flow_from_directory(dataset_train_path, 
                                                                     target_size=target_size,
                                                                     color_mode=color_mode,
                                                                     batch_size=batch_size,
                                                                     class_mode=class_mode,
                                                                     shuffle=shuffle,
                                                                     subset='validation');

                elif(dataset_type == "train-val"):
                    self.system_dict["local"]["data_loaders"]["train"] = self.system_dict["local"]["data_generators"]["train"].flow_from_directory(dataset_train_path, 
                                                                     target_size=target_size,
                                                                     color_mode=color_mode,
                                                                     batch_size=batch_size,
                                                                     class_mode=class_mode,
                                                                     shuffle=shuffle);

                    self.system_dict["local"]["data_loaders"]["val"] = self.system_dict["local"]["data_generators"]["val"].flow_from_directory(dataset_val_path, 
                                                                     target_size=target_size,
                                                                     color_mode=color_mode,
                                                                     batch_size=batch_size,
                                                                     class_mode=class_mode,
                                                                     shuffle=shuffle);


                elif(dataset_type == "csv_train"):
                    train_df, columns = parse_csv2(csv_train, delimiter);

                    self.system_dict["local"]["data_loaders"]["train"] = self.system_dict["local"]["data_generators"]["train"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    subset="training",
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );

                    self.system_dict["local"]["data_loaders"]["val"] = self.system_dict["local"]["data_generators"]["train"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    subset="validation",
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );


                elif(dataset_type == "csv_train-val"):
                    train_df, columns = parse_csv2(csv_train, delimiter);


                    self.system_dict["local"]["data_loaders"]["train"] = self.system_dict["local"]["data_generators"]["train"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );


                    val_df, columns = parse_csv2(csv_val, delimiter);

                    self.system_dict["local"]["data_loaders"]["val"] = self.system_dict["local"]["data_generators"]["val"].flow_from_dataframe(
                                                                    dataframe=val_df,
                                                                    directory=dataset_val_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );

            else:
                if(dataset_type == "csv_train"):
                    train_df, columns = parse_csv2_updated(csv_train, delimiter);

                    self.system_dict["local"]["data_loaders"]["train"] = self.system_dict["local"]["data_generators"]["train"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    subset="training",
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );

                    self.system_dict["local"]["data_loaders"]["val"] = self.system_dict["local"]["data_generators"]["train"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    subset="validation",
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );


                elif(dataset_type == "csv_train-val"):
                    train_df, columns = parse_csv2_updated(csv_train, delimiter);


                    self.system_dict["local"]["data_loaders"]["train"] = self.system_dict["local"]["data_generators"]["train"].flow_from_dataframe(
                                                                    dataframe=train_df,
                                                                    directory=dataset_train_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );


                    val_df, columns = parse_csv2(csv_val, delimiter);

                    self.system_dict["local"]["data_loaders"]["val"] = self.system_dict["local"]["data_generators"]["val"].flow_from_dataframe(
                                                                    dataframe=val_df,
                                                                    directory=dataset_val_path,
                                                                    x_col=columns[0],
                                                                    y_col=columns[1],
                                                                    target_size=target_size,
                                                                    color_mode=color_mode,
                                                                    batch_size=batch_size,
                                                                    class_mode=class_mode,
                                                                    shuffle=shuffle
                                                                );


            self.system_dict["dataset"]["params"]["classes"] = self.system_dict["local"]["data_loaders"]["train"].class_indices;
            self.system_dict["dataset"]["params"]["num_classes"] = len(self.system_dict["dataset"]["params"]["classes"]);

            self.system_dict["dataset"]["params"]["num_train_images"] = len(self.system_dict["local"]["data_loaders"]["train"].labels);
            self.system_dict["dataset"]["params"]["num_val_images"] = len(self.system_dict["local"]["data_loaders"]["val"].labels);

            self.system_dict["dataset"]["status"]= True;
    ###############################################################################################################################################







    ###############################################################################################################################################
    @accepts("self", test=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_dataset_final(self, test=False):
        '''
        Set the transforms and then invoke data loader.

        Args:
            test (bool): If True then test tranforms are set and test dataloader is prepared data, 
                        else train transforms are set and train dataloader is prepared.

        Returns:
            None
        '''
        if(test):
            self.system_dict = set_transform_test(self.system_dict);
        else:
            self.system_dict = set_transform_trainval(self.system_dict);
        self.set_dataset_dataloader(test=test);
    ###############################################################################################################################################

    


from monk.pytorch.testing.imports import *
from monk.system.imports import *

@accepts(str, bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def process_single(img_name, return_raw, system_dict):
    '''
    Run inference on a single image

    Args:
        img_name (str): path to image
        return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                            Else, only the most probable class score is returned back.
                            

    Returns:
        str: predicted class
        float: prediction score
    '''
    img = Image.open(img_name).convert('RGB');
    img = system_dict["local"]["data_transforms"]["test"](img);
    img = img.unsqueeze(0);
    img = Variable(img);
    img = img.to(system_dict["local"]["device"])
    outputs = system_dict["local"]["model"](img)
    l = outputs.data.cpu().numpy().argmax();
    if(system_dict["dataset"]["params"]["classes"]):
        prediction = system_dict["dataset"]["params"]["classes"][l];
    else:
        prediction = l;
    normalized = softmax(outputs.data.cpu().numpy()[0]);
    score = normalized[l];
    if(return_raw):
        return prediction, score, normalized;
    else:
        return prediction, score, "";



@accepts(str, bool, float, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def process_multi(img_name, return_raw, img_thresh, system_dict):
    '''
    Run inference on a single image when label type is multi-label

    Args:
        img_name (str): path to image
        return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                            Else, only the most probable class score is returned back.
        img_thresh (float): Thresholding for multi label image classification.
                            

    Returns:
        list: list of predicted classes
        list: list of predicted scores
    '''
    img = Image.open(img_name).convert('RGB');
    img = system_dict["local"]["data_transforms"]["test"](img);
    img = img.unsqueeze(0);
    img = Variable(img);
    img = img.to(system_dict["local"]["device"])
    outputs = system_dict["local"]["model"](img)
    list_classes = [];
    raw_scores = outputs.cpu().detach().numpy()[0];
    list_scores = [];

    for i in range(len(raw_scores)):
        prob = logistic.cdf(raw_scores[i])
        if(prob > img_thresh):
            list_classes.append(system_dict["dataset"]["params"]["classes"][i])
            list_scores.append(prob)


    if(return_raw):
        return list_classes, list_scores, raw_scores;
    else:
        return list_classes, list_scores, "";


from monk.pytorch.models.imports import *
from monk.system.imports import *
from monk.pytorch.models.common import set_parameter_requires_grad


#classifier 6
set1 = ["alexnet", "vgg11", "vgg11_bn", "vgg13", "vgg13_bn", "vgg16", "vgg16_bn", "vgg19", "vgg19_bn"]

#classifier
set2 = ["densenet121", "densenet161", "densenet169", "densenet201"]

#fc
set3 = ["googlenet", "inception_v3", "resnet18", "resnet34", "resnet50", "resnet101", "resnet152", "resnext50_32x4d", "resnext101_32x8d",
        "shufflenet_v2_x0_5", "shufflenet_v2_x1_0", "shufflenet_v2_x1_5", "shufflenet_v2_x2_0", "wide_resnet101_2", "wide_resnet50_2"]

#classifier 1
set4 = ["mnasnet0_5", "mnasnet0_75", "mnasnet1_0", "mnasnet1_3", "mobilenet_v2", "squeezenet1_0", "squeezenet1_1"]

combined_list = set1+set2+set3+set4
combined_list_lower = list(map(str.lower, combined_list))


@accepts(str, bool, int, bool, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def get_base_model(model_name, use_pretrained, num_classes, freeze_base_network):
    '''
    Get base network for transfer learning based on parameters selected

    Args:
        model_name (str): Select from available models. Check via List_Models() function
        freeze_base_network (bool): If set as True, then base network's weights are freezed (cannot be trained)
        use_gpu (bool): If set as True, uses GPU
        use_pretrained (bool): If set as True, use weights trained on imagenet and coco like dataset
                                Else, use randomly initialized weights.

    Returns:
        neural network: Base network
        str: Name of the model
    '''
    if(model_name not in combined_list_lower):
        print("Model name: {} not found".format(model_name));
    else:
        index = combined_list_lower.index(model_name);
        model_name = combined_list[index];

    if(model_name == "alexnet"):
        finetune_net = torchvision.models.alexnet(pretrained=use_pretrained);
    elif(model_name == "vgg11"):
        finetune_net = torchvision.models.vgg11(pretrained=use_pretrained);
    elif(model_name == "vgg11_bn"):
        finetune_net = torchvision.models.vgg11_bn(pretrained=use_pretrained);
    elif(model_name == "vgg13"):
        finetune_net = torchvision.models.vgg11(pretrained=use_pretrained);
    elif(model_name == "vgg13_bn"):
        finetune_net = torchvision.models.vgg11_bn(pretrained=use_pretrained);
    elif(model_name == "vgg16"):
        finetune_net = torchvision.models.vgg16(pretrained=use_pretrained);
    elif(model_name == "vgg16_bn"):
        finetune_net = torchvision.models.vgg16_bn(pretrained=use_pretrained);
    elif(model_name == "vgg19"):
        finetune_net = torchvision.models.vgg19(pretrained=use_pretrained);
    elif(model_name == "vgg19_bn"):
        finetune_net = torchvision.models.vgg19_bn(pretrained=use_pretrained);
    elif(model_name == "densenet121"):
        finetune_net = torchvision.models.densenet121(pretrained=use_pretrained);
    elif(model_name == "densenet161"):
        finetune_net = torchvision.models.densenet161(pretrained=use_pretrained);
    elif(model_name == "densenet169"):
        finetune_net = torchvision.models.densenet169(pretrained=use_pretrained);
    elif(model_name == "densenet201"):
        finetune_net = torchvision.models.densenet201(pretrained=use_pretrained);
    elif(model_name == "googlenet"):
        finetune_net = torchvision.models.googlenet(pretrained=use_pretrained);
    elif(model_name == "inception_v3"):
        finetune_net = torchvision.models.inception_v3(pretrained=use_pretrained);
    elif(model_name == "resnet18"):
        finetune_net = torchvision.models.resnet18(pretrained=use_pretrained);
    elif(model_name == "resnet34"):
        finetune_net = torchvision.models.resnet34(pretrained=use_pretrained);
    elif(model_name == "resnet50"):
        finetune_net = torchvision.models.resnet50(pretrained=use_pretrained);
    elif(model_name == "resnet101"):
        finetune_net = torchvision.models.resnet101(pretrained=use_pretrained);
    elif(model_name == "resnet152"):
        finetune_net = torchvision.models.resnet152(pretrained=use_pretrained);
    elif(model_name == "resnext50_32x4d"):
        finetune_net = torchvision.models.resnext50_32x4d(pretrained=use_pretrained);
    elif(model_name == "resnext101_32x8d"):
        finetune_net = torchvision.models.resnext101_32x8d(pretrained=use_pretrained);
    elif(model_name == "shufflenet_v2_x0_5"):
        finetune_net = torchvision.models.shufflenet_v2_x0_5(pretrained=use_pretrained);
    elif(model_name == "shufflenet_v2_x1_0"):
        finetune_net = torchvision.models.shufflenet_v2_x1_0(pretrained=use_pretrained);
    elif(model_name == "shufflenet_v2_x1_5"):
        if(use_pretrained):
            msg = "Pretrained model Unavailable for {}.\n".format(model_name);
            msg += "Using xavier initialization";
            ConstraintWarning(msg);
        finetune_net = torchvision.models.shufflenet_v2_x1_5(pretrained=False);
    elif(model_name == "shufflenet_v2_x2_0"):
        if(use_pretrained):
            msg = "Pretrained model Unavailable for {}.\n".format(model_name);
            msg += "Using xavier initialization";
            ConstraintWarning(msg);
        finetune_net = torchvision.models.shufflenet_v2_x2_0(pretrained=False);
    elif(model_name == "wide_resnet101_2"):
        finetune_net = torchvision.models.wide_resnet101_2(pretrained=use_pretrained);
    elif(model_name == "wide_resnet50_2"):
        finetune_net = torchvision.models.wide_resnet50_2(pretrained=use_pretrained);
    elif(model_name == "mnasnet0_5"):
        finetune_net = torchvision.models.mnasnet0_5(pretrained=use_pretrained);
    elif(model_name == "mnasnet0_75"):
        if(use_pretrained):
            msg = "Pretrained model Unavailable for {}.\n".format(model_name);
            msg += "Using xavier initialization";
            ConstraintWarning(msg);
        finetune_net = torchvision.models.mnasnet0_75(pretrained=False);
    elif(model_name == "mnasnet1_0"):
        finetune_net = torchvision.models.mnasnet1_0(pretrained=use_pretrained);
    elif(model_name == "mnasnet1_3"):
        if(use_pretrained):
            msg = "Pretrained model Unavailable for {}.\n".format(model_name);
            msg += "Using xavier initialization";
            ConstraintWarning(msg);
        finetune_net = torchvision.models.mnasnet1_3(pretrained=False);
    elif(model_name == "mobilenet_v2"):
        finetune_net = torchvision.models.mobilenet_v2(pretrained=use_pretrained);
    elif(model_name == "squeezenet1_0"):
        finetune_net = torchvision.models.squeezenet1_0(pretrained=use_pretrained);
    elif(model_name == "squeezenet1_1"):
        finetune_net = torchvision.models.squeezenet1_1(pretrained=use_pretrained);


    finetune_net = set_parameter_requires_grad(finetune_net, freeze_base_network);

    return finetune_net, model_name;




from monk.gluon.finetune.imports import *
from monk.system.imports import *

from monk.gluon.finetune.level_9_transforms_main import prototype_transforms


class prototype_schedulers(prototype_transforms):
    '''
    Main class for all learning rate schedulers in expert mode

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);

    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_fixed(self):
        '''
        Set learning rate fixed

        Args:
            None

        Returns:
            None
        '''
        self.system_dict = scheduler_fixed(self.system_dict);
        
        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################
        

    ###############################################################################################################################################
    @warning_checks(None, None, gamma=["gt", 0.01, "lt", 1], last_epoch=None, post_trace=False)
    @error_checks(None, ["gt", 0], gamma=["gt", 0], last_epoch=None, post_trace=False)
    @accepts("self", int, gamma=float, last_epoch=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_step_decrease(self, step_size, gamma=0.1, last_epoch=-1):
        '''
        Set learning rate to decrease in regular steps

        Args:
            step_size (int): Step interval for decreasing learning rate
            gamma (str): Reduction multiplier for reducing learning rate post every step
            last_epoch (int): Set this epoch to a level post which learning rate will not be decreased

        Returns:
            None
        '''
        self.system_dict = scheduler_step(self.system_dict, step_size, gamma=gamma, last_epoch=last_epoch);
        
        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################
        

    ###############################################################################################################################################
    @warning_checks(None, None, gamma=["gt", 0.01, "lt", 1], last_epoch=None, post_trace=False)
    @error_checks(None, ["inc", None], gamma=["gt", 0], last_epoch=None, post_trace=False)
    @accepts("self", [list, int], gamma=float, last_epoch=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def lr_multistep_decrease(self, milestones, gamma=0.1, last_epoch=-1):
        '''
        Set learning rate to decrease in irregular steps

        Args:
            milestones (list): List of epochs at which learning rate is to be decreased
            gamma (str): Reduction multiplier for reducing learning rate post every step
            last_epoch (int): Dummy variable

        Returns:
            None
        '''
        self.system_dict = scheduler_multistep(self.system_dict, milestones, gamma=gamma, last_epoch=last_epoch);
        
        self.custom_print("Learning rate scheduler");
        self.custom_print("    Name:   {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"]));
        self.custom_print("    Params: {}".format(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]));
        self.custom_print("");
    ###############################################################################################################################################